
/*************************************************************************************
 Arnav Kumar | M259 W23
 
 Project 2: 3D Treemap Visualization
 
 Adapted from Rodger Jieliang Luo's "3D Tree Map Demo" (http://www.mat.ucsb.edu/~g.legrady/academic/courses/21w259/w4/TreeMap_3D_2017.zip), 
 Sara Lafia's "RESEARCH TRAJECTORIES: UCSB Graduate Research Trends" (https://vislab.mat.ucsb.edu/2017/p3/SaraLafia/index.html), 
 Li Zheng's "China Japan 3D Demo" (https://www.mat.ucsb.edu/~g.legrady/academic/courses/23w259/w6/China_Japan_3D.zip), 
 RJ Duran's "Water" (https://vislab.mat.ucsb.edu/2012/p4/rj/index.html), and course references from the MAT259 course.
 
 Data: Seattle Public Library. 
 
 Description: Aggregations of checkouts by year and modality in the Dewey 400-499 categories (spoken and written languages), displayed in treemaps by year.
 
 Range: 2006-2023
 
 Usage: 
 1. Enter a modality number from 1 to 46 to highlight respective modalities.
 2. Press 'p' to toggle presentation mode
 3. Press Spacebar to cycle through treemap modalities
 4. Press 't' to cycle treemap visual transparency
 5. Press 'b' to cycle background colors
 6. Press 'd' to to cycle through gradient line draw speeds
 7. Press 'f' to to cycle through fields of view
 8. Press '+' to capture and save a screenshot of the treemap
 9. Navigate Peasy Cam to explore the environment. 
 
 *************************************************************************************/

import treemap.*; //Tree Map library
import peasy.*; //3D Navigation library
import controlP5.*;

//Constants Section
final int NUM_LAYERS = 18;   //Number of Layers
int LAYER_DIST = 200;  //Distance between each layer
int TRANS = 50;

int selectedmodality = -100;

boolean presentationMode = true;

int background = 0;

int drawspeed = 10;

float fov;

float cameraZ;

//Camera(3D Nav) varaible
PeasyCam cam;

ControlP5 cp5;

//Data-Related varaibles
Table table;
int numRows, numColumns;
float[][] dataMatrix;
String[] modalities;

//TreeMap-Related varaibles
modalityMap[] modalityData;
Treemap[] map;

ArrayList<ArrayList<PVector>> pointsOne = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsTwo = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsThree = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsFour = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);

void setup() {
  smooth();
  
  size(1440, 820, OPENGL);
  
  // fullScreen(OPENGL);
  
  cam = new PeasyCam(this, 4000);
  
  cam.setMinimumDistance(1);
  
  cam.setMaximumDistance(4000);

  cp5 = new ControlP5(this);
  
  cp5.addTextarea("txt")
    .setPosition(0,40)
    .setSize(310,320)
    .setLineHeight(14)
    //.setColor(color(200))
    .setColorBackground(color(10, 100))
    .setColorForeground(color(10, 100));
    ;
    
  cp5.addTextarea("title")
    .setPosition(0, 0)
    .setSize(950, 40)
    .setLineHeight(14)
    .setColor(color(200))
    .setColorBackground(color(10, 100))
    .setColorForeground(color(10, 100));
    ;
    
  cp5.addTextfield("modality")
    .setPosition(20, 50)
    .setSize(50, 20)
    .setFocus(true)
    .setColor(color(0))
    .setColorBackground(color(255))
    .setCaptionLabel(" ")
    ;

  cp5.addTextlabel("label")
    .setText("Please Enter A Modality Number from 1 to 46")
    .setPosition(15, 90)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
    
  /* cp5.addCheckBox("checkBox")
    .setPosition(20, 150)
    .setSize(25, 25) 
    .addItem(" ", 1)
    .activateAll()
    ; */
      
   cp5.addTextlabel("p")
    .setText("Press 'p' to toggle presentation mode")
    .setPosition(15, 160)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
 
   cp5.addTextlabel("Spacebar")
    .setText("Press Spacebar to cycle through treemap modalities")
    .setPosition(15, 190)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
    
   cp5.addTextlabel("t")
    .setText("Press 't' to cycle treemap visual transparency")
    .setPosition(15, 220)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
    
 cp5.addTextlabel("b")
    .setText("Press 'b' to cycle background colors")
    .setPosition(15, 250)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
    
 cp5.addTextlabel("d")
    .setText("Press 'd' to to cycle through gradient line draw speeds")
    .setPosition(15, 280)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
    
 /* cp5.addTextlabel("f")
    .setText("Press 'f' to cycle through gradient line draw speeds")
    .setPosition(15, 170)
    .setColorValue(0)
    .setFont(createFont("Arial", 10))
    ;
    */
    
  cp5.addTextlabel("f")
    .setText("Press 'f' to to cycle through fields of view")
    .setPosition(15, 310)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
    
  cp5.addTextlabel("+")
    .setText("Press '+' to capture and save a screenshot of the treemap")
    .setPosition(15, 340)
    .setColorValue(0)
    .setFont(createFont("Helvetica", 10))
    ;
    
  cp5.setAutoDraw(false);
  // Set the angle of view - like changing the lens on a camera PI/2 (wide angle); PI/6 normal lens; PI/12 (telephoto)
  fov = (PI / 10) + (PI / 4); // field of view try between PI/2 to PI/10
  cameraZ = (height/2.0) / tan(PI/6);
  //perspective(field of view in radians; aspect/ratio of width to height; zNear - Z position of nearest clipping, zFar - Z position of farthest clipping)
  //default values are : perspective(PI/3.0, width/height, cameraZ/10.0, cameraZ*10.0) where cameraZ is ((height/2.0) / tan(PI*60.0/360.0));
  perspective(fov, float(width)/float(height), cameraZ/10.0, cameraZ*10.0); // if using 0.001 instead of cameraZ/10.0 so that all things close to the screen can be seen - get noise

  //Put data into Processing table from csv
  table = loadTable("ArnavKumarMAT259Project2ExpandedQuery.csv", "header");
  numRows = table.getRowCount(); 
  //numRows = 1;
  numColumns = table.getColumnCount();
  dataMatrix = new float[numColumns-1][numRows];
  modalities = new String[numRows];
  println("Rows: " + numRows + " Columns: " + numColumns);

  //Put data from Table to 2D array
  for (int i=0; i<numColumns-1; i++) {
    for (int j=0; j<numRows; j++) {
      dataMatrix[i][j] = table.getFloat(j, i+1);
    }
  }

  //Save modality categories into 1D array
  for (int i=0; i<numRows; i++) {
    modalities[i] = table.getString(i, 0);
  }

  //Init tree map varaibles
  modalityData = new modalityMap[NUM_LAYERS];
  map = new Treemap[NUM_LAYERS];
  for (int i=0; i<NUM_LAYERS; i++) { 
    modalityData[i]= new modalityMap();
    pointsOne.add(new ArrayList<PVector>());
    pointsTwo.add(new ArrayList<PVector>());
    pointsThree.add(new ArrayList<PVector>());
    pointsFour.add(new ArrayList<PVector>());
  }

  //Add modality categories and counts into tree map data structure
  for (int i=0; i<dataMatrix.length; i++) {
    for (int j=0; j<dataMatrix[i].length; j++) {
      modalityData[i].addmodality(modalities[j], (int)dataMatrix[i][j]);
    }
    modalityData[i].finishAdd();
    map[i] = new Treemap(modalityData[i], 0, 0, width, height);
  }

  //Draw basic tree maps 
  for (int i=0; i<dataMatrix.length; i++) {
    //Find and save the central point of each rect
    for (int j=0; j<modalityData[i].getItems().length; j++) {
      double x = modalityData[i].getItems()[j].getBounds().x;
      double y = modalityData[i].getItems()[j].getBounds().y;
      double w = modalityData[i].getItems()[j].getBounds().w; 
      double h = modalityData[i].getItems()[j].getBounds().h; 
      //PVector centralPoint = new PVector((int)(x+w/2), (int)(y+h/2), LAYER_DIST*i);
      //centralPoints.get(i).add(centralPoint);
      PVector pointOne = new PVector((int)x, (int)y, LAYER_DIST*(i-dataMatrix.length/2));
      pointsOne.get(i).add(pointOne);

      PVector pointTwo = new PVector((int)(x+w), (int)y, LAYER_DIST*(i-dataMatrix.length/2));
      pointsTwo.get(i).add(pointTwo);

      PVector pointThree = new PVector((int)(x+w), (int)(y+h), LAYER_DIST*(i-dataMatrix.length/2));
      pointsThree.get(i).add(pointThree);

      PVector pointFour = new PVector((int)x, (int)(y+h), LAYER_DIST*(i-dataMatrix.length/2));
      pointsFour.get(i).add(pointFour);
    }
  }
}

void draw() {
  gui();
  
  frameRate(drawspeed);
  colorMode(HSB, 360, 100, 100, 100);
  background((background / 2), background, 255);
  
  if (cp5.getWindow().isMouseOver()) {
    cam.setActive(false);
  } else {
    cam.setActive(true);
  }

  if (presentationMode) {
    rotateY(frameCount*0.001);
    if (frameCount % 120 == 0) {
      selectedmodality += 1;
      if (selectedmodality == 46) selectedmodality = -1;
    }
  }

  //Draw Top
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsOne.size()-1; i++) {
     float x1_1 = pointsOne.get(i).get(j).x;
     float y1_1 = pointsOne.get(i).get(j).y + 0.1;
     float z1_1 = pointsOne.get(i).get(j).z;

     float x2_1 = pointsTwo.get(i).get(j).x;
     float y2_1 = pointsTwo.get(i).get(j).y + 0.1;
     float z2_1 = pointsTwo.get(i).get(j).z;

     float x1_2 = pointsOne.get(i+1).get(j).x;
     float y1_2 = pointsOne.get(i+1).get(j).y + 0.1;
     float z1_2 = pointsOne.get(i+1).get(j).z;

     float x2_2 = pointsTwo.get(i+1).get(j).x;
     float y2_2 = pointsTwo.get(i+1).get(j).y + 0.1;
     float z2_2 = pointsTwo.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(modalityData[i].modalities.keySet().toArray()[j].toString());

     beginShape();  
     noStroke();
     if (lineColor == selectedmodality) {
       fill(map(lineColor, 0, 46, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedmodality >= 0) noFill();
       else fill(map(lineColor, 0, 46, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Bottom
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsThree.size()-1; i++) {
     float x1_1 = pointsThree.get(i).get(j).x;
     float y1_1 = pointsThree.get(i).get(j).y - 0.1;
     float z1_1 = pointsThree.get(i).get(j).z;

     float x2_1 = pointsFour.get(i).get(j).x;
     float y2_1 = pointsFour.get(i).get(j).y - 0.1;
     float z2_1 = pointsFour.get(i).get(j).z;

     float x1_2 = pointsThree.get(i+1).get(j).x;
     float y1_2 = pointsThree.get(i+1).get(j).y - 0.1;
     float z1_2 = pointsThree.get(i+1).get(j).z;

     float x2_2 = pointsFour.get(i+1).get(j).x;
     float y2_2 = pointsFour.get(i+1).get(j).y - 0.1;
     float z2_2 = pointsFour.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(modalityData[i].modalities.keySet().toArray()[j].toString());     

     beginShape();
     noStroke();
     if (lineColor == selectedmodality) {
       fill(map(lineColor, 0, 46, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedmodality >= 0) noFill();
       else fill(map(lineColor, 0, 46, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Left
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsOne.size()-1; i++) {
     float x1_1 = pointsOne.get(i).get(j).x + 0.1;
     float y1_1 = pointsOne.get(i).get(j).y;
     float z1_1 = pointsOne.get(i).get(j).z;

     float x2_1 = pointsFour.get(i).get(j).x + 0.1;
     float y2_1 = pointsFour.get(i).get(j).y;
     float z2_1 = pointsFour.get(i).get(j).z;

     float x1_2 = pointsOne.get(i+1).get(j).x + 0.1;
     float y1_2 = pointsOne.get(i+1).get(j).y;
     float z1_2 = pointsOne.get(i+1).get(j).z;

     float x2_2 = pointsFour.get(i+1).get(j).x + 0.1;
     float y2_2 = pointsFour.get(i+1).get(j).y;
     float z2_2 = pointsFour.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(modalityData[i].modalities.keySet().toArray()[j].toString());     

     beginShape();
     noStroke();
     if (lineColor == selectedmodality) {
       fill(map(lineColor, 0, 46, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedmodality >= 0) noFill();
       else fill(map(lineColor, 0, 46, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Right
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsTwo.size()-1; i++) {
     float x1_1 = pointsTwo.get(i).get(j).x - 0.1;
     float y1_1 = pointsTwo.get(i).get(j).y;
     float z1_1 = pointsTwo.get(i).get(j).z;

     float x2_1 = pointsThree.get(i).get(j).x - 0.1;
     float y2_1 = pointsThree.get(i).get(j).y;
     float z2_1 = pointsThree.get(i).get(j).z;

     float x1_2 = pointsTwo.get(i+1).get(j).x - 0.1;
     float y1_2 = pointsTwo.get(i+1).get(j).y;
     float z1_2 = pointsTwo.get(i+1).get(j).z;

     float x2_2 = pointsThree.get(i+1).get(j).x - 0.1;
     float y2_2 = pointsThree.get(i+1).get(j).y;
     float z2_2 = pointsThree.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(modalityData[i].modalities.keySet().toArray()[j].toString());     
     lineColor = map(lineColor, 0, 46, 0, 255);

     beginShape();
     noStroke();
     if (lineColor == selectedmodality) {
       fill(map(lineColor, 0, 46, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedmodality >= 0) noFill();
       else fill(map(lineColor, 0, 46, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Draw basic tree maps 
  for (int i=0; i<dataMatrix.length; i++) {
    pushMatrix();
    translate(-width/2, -height/2, LAYER_DIST*(i - dataMatrix.length/2));
    map[i].draw();
    popMatrix();
  }

  //Draw Title
  cam.beginHUD();  //Stopping peasy cam
  textSize(15);
  fill(0);
  text("Time of Change Across Modalities of Language Learning (2006-2023 at the Seattle Public Library) | A 3D Tree Map Collection by Arnav Kumar", 480, 20);
  cp5.draw();
  cam.endHUD();
}

public void modality(String theText) {
  // automatically receives results from controller input
  println("a textfield event for controller 'input' : "+theText);
  selectedmodality = int(theText);
}

/* void checkBox(float[] a) {
  if(a[0] == 1) presentationMode = true;
  else presentationMode = false;
} */

void gui() {
  hint(DISABLE_DEPTH_TEST);
  cam.beginHUD();
  cp5.draw();
  cam.endHUD();
  hint(ENABLE_DEPTH_TEST);
}
